#include "NMenu.h"

#include <Message.h>
#include <Handler.h>
#include <Bitmap.h>
#include <Window.h>
#include <Font.h>
#include <Screen.h>

#include <stdio.h>

// some nice bulky function to draw a frame somewhat as seen in the GONX screenshots
void GonxFrame( BView *view, BRect fullrect, rgb_color backcolor )
{
	BRect rect = fullrect;
	rect.left += 1;
	rect.top += 1;
	rect.right -= 1;
	rect.bottom -= 1;

	view->SetDrawingMode( B_OP_ALPHA );

	view->SetHighColor( 207, 204, 195, 150 );
	view->StrokeRoundRect( fullrect, 5, 5 );

	view->SetDrawingMode( B_OP_COPY );

	BRect r = rect;
	r.left += 2;
	r.top += 2;
	r.right -= 2;
	r.bottom -= 2;
	view->SetHighColor( backcolor );

	if ( backcolor.alpha < 255 )
		view->SetDrawingMode( B_OP_ALPHA );

	view->FillRect( r );

	view->StrokeLine( BPoint( r.left-1, r.top-1 ), BPoint( r.right, r.top-1) );
	view->StrokeLine( BPoint( r.left-1, r.top-1 ), BPoint( r.left-1, r.bottom) );
	view->StrokeLine( BPoint( r.right+1, r.top-1 ), BPoint( r.right+1, r.bottom) );
	view->StrokeLine( BPoint( r.left-1, r.bottom+1 ), BPoint( r.right, r.bottom+1) );

	view->SetDrawingMode( B_OP_ALPHA );
	view->SetHighColor( 255, 255, 255, 100 );
	view->StrokeLine( BPoint( r.left-1, r.top-1 ), BPoint( r.right, r.top-1) );
	view->StrokeLine( BPoint( r.left-1, r.top-1 ), BPoint( r.left-1, r.bottom) );
	view->StrokeLine( BPoint( r.right+1, r.top-1 ), BPoint( r.right+1, r.bottom) );
	view->StrokeLine( BPoint( r.left-1, r.bottom+1 ), BPoint( r.right, r.bottom+1) );

	view->SetDrawingMode( B_OP_COPY );
	view->SetHighColor( 207, 204, 195 );
	view->StrokeRoundRect( rect, 3, 3 );
}

// gets the string height from a view's font, or from be_plain_font
// if the view is null
static inline float StringHeight( BView *view )
{
	font_height fontHeight;
	const BFont *font;
	BFont _font;
	if ( !view )
		font = be_plain_font;
	else
	{
		view->GetFont( &_font );
		font = &_font;
	}
	font->GetHeight(&fontHeight);
	float fTitleAscent = ceilf(fontHeight.ascent);
	float fTitleDescent = ceilf(fontHeight.descent + fontHeight.leading);
	return fTitleAscent + fTitleDescent;
}

TMenuItem::TMenuItem( const char *label, BMessage *message )
	: fMessage( message ), fMenu(0), fSubmenu(0), fMarked(false), fEnabled(true), fSelected(false)
{
	fDraggable = false;

	SetLabel( label );
}

TMenuItem::TMenuItem( TMenu *submenu, BMessage *message )
	: fMessage( message ), fMenu(0), fSubmenu( submenu ), fMarked(false), fEnabled(true), fSelected(false)
{
	fDraggable = false;

	SetLabel( submenu->Name() );
}

TMenuItem::~TMenuItem()
{
	if ( fMessage )
		delete fMessage;
	if ( fSubmenu )
		delete fSubmenu;
}

BPoint TMenuItem::ContentLocation() const
{
	return fFrame.LeftTop();
}

void TMenuItem::Draw()
{
	if ( fSelected && IsEnabled() )
		DrawHighlightFrame();
	fMenu->MovePenTo( BPoint( fFrame.left+16, fFrame.top ) );
	DrawContent();
}

void TMenuItem::DrawContent()
{
	BPoint where = fMenu->PenLocation();
	where.y += fFrame.Height() / 2 + 4;

	fMenu->SetHighColor( 0, 0, 0 );

	// todo: cache this
	char newlabel[256];
	TruncateLabel( fFrame.Width() - 32, newlabel );

	if ( !IsEnabled() )
		fMenu->SetHighColor( 130, 130, 130 );
	else
		fMenu->SetHighColor( 0, 0, 0 );

	if ( fSelected && fEnabled )
		fMenu->SetLowColor( 210, 210, 193 );
	else
		fMenu->SetLowColor( 239, 235, 231 );

	fMenu->DrawString( newlabel, where );
}

BRect TMenuItem::Frame() const
{
	return BRect( 0, 0, fFrame.Width(), fFrame.Height() );
}

void TMenuItem::GetContentSize( float *width, float *height )
{
	float _width = fMenu->StringWidth( fLabel.String() );
	float _height = StringHeight( fMenu );

	// some aditional space
	_width += 32;
	_height += 8;

	// todo: cache these values

	*width = _width;
	*height = _height;
}

void TMenuItem::Highlight( bool flag )
{
	fSelected = flag;
	fMenu->Invalidate( fFrame );
}

status_t TMenuItem::Invoke( BMessage *message )
{
	if ( !message )
		message = fMessage;

	if ( message )
	{
		// if the message already has any of this fields, replace them
		if ( message->ReplacePointer( "source", this ) != B_OK )
			message->AddPointer( "source", this );
		int64 now = real_time_clock_usecs();
		if ( message->ReplaceInt64( "when", now ) != B_OK )
			message->AddInt64( "when", now );
		int32 index = fMenu->IndexOf(this);
		if ( message->ReplaceInt32( "index", index ) != B_OK )
			message->AddInt32( "index", index );
	}

	return B_OK;
}

TMenu *TMenuItem::Menu() const
{
	return fMenu;
}

bool TMenuItem::IsSelected() const
{
	return fSelected;
}

void TMenuItem::SetEnabled( bool enabled )
{
	fEnabled = enabled;
	if ( fMenu )
		fMenu->Invalidate( fFrame );
}

bool TMenuItem::IsEnabled() const
{
	if ( !fEnabled )
		return false;
	if ( fMenu && !fMenu->IsEnabled() )
		return false;
	return true;
}

void TMenuItem::SetLabel( const char *string )
{
	fLabel = string;
}

const char *TMenuItem::Label() const
{
	return fLabel.String();
}

void TMenuItem::SetMarked( bool flag )
{
	fMarked = flag;
}

bool TMenuItem::IsMarked() const
{
	return fMarked;
}

void TMenuItem::SetShortcut( char shortcut, uint32 modifiers )
{
	// todo
}

char TMenuItem::Shortcut( uint32 *modifiers ) const
{
	// todo
	return -1;
}

void TMenuItem::SetTrigger( char trigger )
{
	// todo
}

char TMenuItem::Trigger() const
{
	// todo
	return -1;
}

TMenu *TMenuItem::Submenu() const
{
	return fSubmenu;
}

void TMenuItem::TruncateLabel( float maxWidth, char *newLabel )
{
	BFont font;
	fMenu->GetFont( &font );

	BString string( fLabel );
	font.TruncateString( &string, B_TRUNCATE_SMART, maxWidth - 8);
	strcpy( newLabel, string.String() );
}

BMessage *TMenuItem::DragMessage()
{
	return 0;
}

void TMenuItem::DrawHighlightFrame()
{
	GonxFrame( fMenu, fFrame, (rgb_color) { 210, 210, 193, 255 } );
}

TSeparatorItem::TSeparatorItem()
	: TMenuItem( (const char *)0 )
{
	SetEnabled(false);
}

void TSeparatorItem::SetEnabled( bool state )
{
	TMenuItem::SetEnabled(false);
}

void TSeparatorItem::GetContentSize( float *width, float *height )
{
	*width = 0;
	*height = 8;
}

void TSeparatorItem::Draw()
{
	BPoint p = ContentLocation();
	BRect frame = Frame();
	BRect r( p.x, p.y, p.x+frame.right, p.y+frame.bottom );
	Menu()->SetHighColor( 207, 204, 195 );
	Menu()->StrokeLine( BPoint( r.left, r.top + 3), BPoint( r.right, r.top +3 ) );
	Menu()->StrokeLine( BPoint( r.left, r.bottom - 3), BPoint( r.right, r.bottom -3 ) );
}

TBitmapMenuItem::TBitmapMenuItem( const char *label, BBitmap *bitmap, BMessage *message )
	: TMenuItem( label, message ), fBitmap( bitmap ), fBitmapAutoDestruct( true )
{
}

TBitmapMenuItem::TBitmapMenuItem( TMenu *submenu, BBitmap *bitmap, BMessage *message )
	: TMenuItem( submenu, message ), fBitmap( bitmap ), fBitmapAutoDestruct( true )
{
}

TBitmapMenuItem::~TBitmapMenuItem()
{
	if ( fBitmapAutoDestruct )
		delete fBitmap;
}

void TBitmapMenuItem::GetContentSize( float *width, float *height )
{
	TMenuItem::GetContentSize( width, height );
	if ( fBitmap )
	{
		*width += fBitmap->Bounds().Width() + 8;
		if ( *height < (fBitmap->Bounds().Height()+8) )
			*height = fBitmap->Bounds().Height()+8;
	}
}

void TBitmapMenuItem::DrawContent()
{
	if ( fBitmap )
	{
		BPoint p = ContentLocation();

		TMenu *menu = Menu();

		menu->SetDrawingMode( B_OP_ALPHA );
		if ( !IsEnabled() )
		{
			menu->SetBlendingMode( B_CONSTANT_ALPHA, B_ALPHA_OVERLAY );
			menu->SetHighColor( 255, 255, 255, 100 );
		}
		else
			menu->SetBlendingMode( B_PIXEL_ALPHA, B_ALPHA_OVERLAY );

		BPoint where( p.x + 8 + 4, (Frame().Height() - fBitmap->Bounds().Height())/2 + p.y);
		if ( fBitmap->Bounds().Width() > 16 )
			where.x -= 8;

		menu->DrawBitmap( fBitmap, where );
		menu->MovePenTo( p.x+fBitmap->Bounds().Width()+where.x+8, p.y );

		menu->SetDrawingMode( B_OP_COPY );
	}
	TMenuItem::DrawContent();
}

void TBitmapMenuItem::TruncateLabel( float maxWidth, char *newLabel )
{
	TMenuItem::TruncateLabel( maxWidth - ( fBitmap ? ( fBitmap->Bounds().Width() - 8 ) : 0 ), newLabel );
}

void TBitmapMenuItem::SetBitmapAutoDestruct( bool state )
{
	fBitmapAutoDestruct = state;
}

bool TBitmapMenuItem::BitmapAutoDestruct() const
{
	return fBitmapAutoDestruct;
}

TMenu::TMenu( const char *name )
	: BView( BRect( 0, 0, 0, 0 ), name, 0, B_WILL_DRAW )
{
	fBackgroundColor = (rgb_color) { 239, 235, 231, 225 };
	// warning: maybe shouldnt use B_TRANSPARENT_COLOR.. but with B_TRANSPARENT_COLOR we get much less flicker
	SetViewColor( B_TRANSPARENT_COLOR );

	fAlphedBack = 0;

	fHighlightedItem = 0;
	fHighlightedItemIndex = 0;

	fSuperItem = 0;

	fUseAlphaBlending = false;

	SetItemMargins( 2, 2, 2, 2 );

	fMaxContentWidth = 500;

	fRadioMode = true;
	fEnabled = true;

#ifdef B_BEOS_VERSION_DANO
	SetDoubleBuffering( 1 );
#endif
}

TMenu::~TMenu()
{
	int32 count = fItems.CountItems();

	for ( int32 i = 0; i<count; i++ )
	{
		TMenuItem *item = static_cast<TMenuItem*>( fItems.ItemAt(i) );
		delete item;
	}

	if ( fAlphedBack )
		delete fAlphedBack;
}

bool TMenu::AddItem( TMenuItem *item )
{
	return AddItem( item, fItems.CountItems() );
}

bool TMenu::AddItem( TMenuItem *item, int32 index )
{
	item->fMenu = this;

	if ( fItems.AddItem( item, index ) )
	{
//		todo: check this, when adding an item, should invalidate the layout
//		InvalidateSizes( item );
		return true;
	}
	else
		return false;
}

bool TMenu::AddItem( TMenu *submenu )
{
	return AddItem( submenu, fItems.CountItems() );
}

bool TMenu::AddItem( TMenu *submenu, int32 index )
{
	TMenuItem *item = new TMenuItem( submenu );
	return AddItem( item, index );
}

bool TMenu::AddSeparatorItem()
{
	TSeparatorItem *item = new TSeparatorItem();
	if ( !item )
		return false;
	return AddItem( item );
}

void TMenu::AttachedToWindow()
{
}

int32 TMenu::CountItems() const
{
	return fItems.CountItems();
}

void TMenu::Draw( BRect updateRect )
{
	if ( fAlphedBack )
		DrawBitmap( fAlphedBack, updateRect, updateRect );
	if ( fUseAlphaBlending )
	{
		SetDrawingMode( B_OP_ALPHA );
	}

	rgb_color bc = fBackgroundColor;
	if ( !fUseAlphaBlending )
		bc.alpha = 255;
	GonxFrame( this, Bounds(), bc );

	PushState();

	// todo: of course this isnt optimised at all
	int32 count = CountItems();
	for ( int32 i=0; i<count; i++ )
	{
		TMenuItem *item = static_cast<TMenuItem*>(fItems.ItemAt(i));
		if ( item->fFrame.Intersects( updateRect ) )
			item->Draw();
	}

	PopState();
}

TMenuItem *TMenu::FindItem( const char *label ) const
{
	int32 count = fItems.CountItems();

	for ( int32 i = 0; i<count; i++ )
	{
		TMenuItem *item = static_cast<TMenuItem*>( fItems.ItemAt(i) );
		if ( !strcmp( item->Label(), label ) )
			return item;
	}

	return 0;
}

TMenuItem *TMenu::FindItem( uint32 command ) const
{
	int32 count = fItems.CountItems();

	for ( int32 i = 0; i<count; i++ )
	{
		TMenuItem *item = static_cast<TMenuItem*>( fItems.ItemAt(i) );
		if ( item->fMessage && item->fMessage->what == command )
			return item;
	}

	return 0;
}

TMenuItem *TMenu::FindMarked()
{
	// todo
	return 0;
}

void TMenu::Hide()
{
	if ( Window() )
		Window()->Hide();
}

void TMenu::Show( bool selectFirst )
{
	// todo
}

void TMenu::Show()
{
}

int32 TMenu::IndexOf( TMenuItem *item ) const
{
	return fItems.IndexOf( item );
}

int32 TMenu::IndexOf( TMenu *submenu ) const
{
	int32 count = fItems.CountItems();

	for ( int32 i = 0; i<count; i++ )
	{
		TMenuItem *item = static_cast<TMenuItem*>( fItems.ItemAt(i) );
		if ( item->fSubmenu == submenu )
			return i;
	}

	return -1;
}

void TMenu::InvalidateLayout()
{
	InvalidateSizes();
}

TMenuItem *TMenu::ItemAt( int32 index ) const
{
	return static_cast<TMenuItem*>(fItems.ItemAt( index ));
}

TMenu *TMenu::SubmenuAt( int32 index ) const
{
	TMenuItem *item = static_cast<TMenuItem*>(fItems.ItemAt( index ));
	if ( item )
		return item->fSubmenu;
	return 0;
}

void TMenu::KeyDown( const char *bytes, int32 numBytes )
{
	// todo: write this
}

TMenuItem *TMenu::RemoveItem( int32 index )
{
	TMenuItem *item = static_cast<TMenuItem*>( fItems.RemoveItem( index ) );
	if ( item )
	{
		TMenuItem *item2 = ItemAt( index + 1 );
		InvalidateSizes( item2 );
	}
	return item;
}

bool TMenu::RemoveItem( TMenuItem *item )
{
	bool res = fItems.RemoveItem( item );
	if ( res )
	{
		int32 index = IndexOf( item );
		TMenuItem *item2 = ItemAt( index + 1 );
		InvalidateSizes( item2 );
	}
	return res;
}

bool TMenu::RemoveItem( TMenu *submenu )
{
	int32 index = IndexOf( submenu );
	if ( index == -1 )
		return false;
	return RemoveItem( ItemAt( index ) );
}

BPoint TMenu::ScreenLocation()
{
	return Window()->Frame().LeftTop();
}

void TMenu::SetEnabled( bool enabled )
{
	fEnabled = enabled;
}

bool TMenu::IsEnabled() const
{
	return fEnabled;
}

void TMenu::SetItemMargins( float left, float top, float right, float bottom )
{
	fItemMargins = BRect (left, top, right, bottom );
	InvalidateSizes();
}

void TMenu::GetItemMargins( float *left, float *top, float *right, float *bottom )
{
	*left = fItemMargins.left;
	*top = fItemMargins.top;
	*right = fItemMargins.right;
	*bottom = fItemMargins.bottom;
}

void TMenu::SetLabelFromMarked( bool flag )
{
	// todo
}

bool TMenu::IsLabelFromMarked()
{
	return false;
}

void TMenu::SetMaxContentWidth( float width )
{
	fMaxContentWidth = width;
	// todo: check this, we should invalidate layout here
}

float TMenu::MaxContentWidth() const
{
	return fMaxContentWidth;
}

void TMenu::SetRadioMode( bool flag )
{
	fRadioMode = flag;
}

bool TMenu::IsRadioMode()
{
	return fRadioMode;
}

status_t TMenu::SetTargetForItems( BHandler *handler )
{
	return SetTargetForItems( BMessenger(handler) );
}

status_t TMenu::SetTargetForItems( BMessenger messenger )
{
	fTarget = messenger;
	return fTarget.IsValid() ? B_OK : B_ERROR;
}

void TMenu::SetTriggersEnabled( bool flag )
{
	// todo
}

bool TMenu::AreTriggersEnabled() const
{
	return false;
}

TMenuItem *TMenu::Superitem() const
{
	return fSuperItem;
}

TMenu *TMenu::Supermenu() const
{
	if ( fSuperItem )
		return fSuperItem->fMenu;
	return 0;
}

namespace TPrivate {

// todo: check this, maybe a popup window should be reusable
class TSubmenuPopup : public BWindow
{
public:
	TSubmenuPopup( TMenu *submenu, BPoint where )
		: BWindow( BRect( 0, 0, 200, 200 ), submenu->Name(), B_NO_BORDER_WINDOW_LOOK, B_FLOATING_ALL_WINDOW_FEEL, B_NOT_CLOSABLE ),
			fMenu( submenu ), fSubmenu(0)
	{
		AddChild( submenu );
		MoveTo( where );
	}

	~TSubmenuPopup()
	{
		// don't delete the TMenu
		RemoveChild( fMenu );

		if ( fSubmenu )
		{
			fSubmenu->Lock();
			fSubmenu->Quit();
		}
	}

	void Go( bool mode = true)
	{
		fMenu->Show();

		fMenu->InvalidateSizes();
		ResizeTo( fMenu->Bounds().Width(), fMenu->Bounds().Height() );

		BScreen screen(this);
		BRect screenframe = screen.Frame();
		BPoint where = Frame().LeftTop();
		if ( Frame().bottom > screenframe.bottom ) // the menu is going to out of the screen
		{
			if ( !mode )
				where.y -= Bounds().Height();
			else
				where.y = screenframe.bottom - Bounds().Height();
		}
		if ( Frame().right > screenframe.right )
		{
			if (fParent)
				where.x = fParent->Frame().left - Bounds().Width() + 5;
			else
				where.x -= Bounds().Width();
		}

		MoveTo( where );

		fMenu->PrepareAlphaBlending();

		Show();
	}

	TMenuItem *SingleTrack( BPoint where, TSubmenuPopup **target )
	{
		Lock();
		BPoint point( where );
		fMenu->ConvertFromScreen( &point );
		Unlock();
		TMenuItem *item = 0;
		if ( fSubmenu )
			item = fSubmenu->SingleTrack( where, target );
		if ( !item )
		{
			item = fMenu->ItemAtPoint( point );
			if ( item )
				*target = this;
		}
		return item;
	}

	void CloseAllSubMenus()
	{
		if ( fSubmenu )
		{
			fSubmenu->Lock();
			fSubmenu->Quit();
			fSubmenu = 0;
		}
	}

	bool IsSubmenu( TSubmenuPopup *menu )
	{
		if ( !menu || !fMenu )
			return false;
		if ( menu == fSubmenu )
			return true;
		if ( fSubmenu )
			return fSubmenu->IsSubmenu(menu);
		return false;
	}

	TMenu *fMenu;
	TSubmenuPopup *fSubmenu, *fParent;
};

};

// todo: check this, openAnyway still doesn't work as it should
//			the menu item is being invoked on mouse down and not on
//			mouse up as expected
TMenuItem *TMenu::Track( bool openAnyway, BRect *clickToOpenRect )
{
	BPoint point, prev_point;
	uint32 buttons, prev_buttons;
	bool open = false;
	TPrivate::TSubmenuPopup *popup = static_cast<TPrivate::TSubmenuPopup*>( Window() );
	TPrivate::TSubmenuPopup *prev_target = 0, *last_valid_target = 0;
	TMenuItem *phitem = 0, *last_launching_item = 0;
	int32 times_in_this_item = 0;

	snooze(250000);

	Window()->Lock();
	GetMouse( &point, &buttons );
	Window()->Unlock();

	if ( !buttons && !openAnyway )
		return 0;

	if ( clickToOpenRect && openAnyway )
	{
		if ( !clickToOpenRect->Contains( point ) )
			return 0;
	}

	open = !buttons;

	while ( 1 )
	{
		prev_point = point;
		prev_buttons = buttons;
		Window()->Lock();
		GetMouse( &point, &buttons );
		TMenuItem *item = 0;
		TPrivate::TSubmenuPopup *target = 0;
		ConvertToScreen( &point );
		Window()->Unlock();
		item = popup->SingleTrack( point, &target );

#if 0 // todo: dragging
		if ( open && buttons )
		{
			if ( item && item->fDraggable )
			{
				int count = 5;
				while ( count )
				{
					count --;
					BPoint pw;
					uint32 buts;
					Window()->Lock();
					GetMouse( &pw, &buts );
					Window()->Unlock();
					if ( !buts )
						break;
					if ( pw != point )
					{
						BMessage *message = item->DragMessage();
						Window()->Lock();
						DragMessage( message, item->fFrame );
						Window()->Unlock();
						delete message;
						return 0;
					}
					snooze( 500000 / 5 );
				}
			}
		}
#endif

		if ( (!buttons && !open) || (buttons && open) )
		{
			if ( item && item->fEnabled )
			{
				if ( !fRadioMode )
					item->SetMarked( !item->fMarked );
				return item;
			}
			else
			{
				return 0;
			}
		}
		else
		{
			if ( !item )
			{
				if ( last_valid_target )
					last_valid_target->CloseAllSubMenus();
			}

			if ( (item != phitem) )
			{
				times_in_this_item = 0;

				// we setup fHighlightedItem because it helps the menu to better find the menu item that was hit
				if ( phitem )
				{
					phitem->fMenu->fHighlightedItem = 0;
					phitem->fMenu->fHighlightedItemIndex = -1;
					if ( prev_target->Lock() )
					{
						phitem->Highlight( false );
						prev_target->Unlock();
					}
				}

				if ( item )
				{
					if ( last_launching_item && last_launching_item != item )
					{
						TPrivate::TSubmenuPopup *father = static_cast<TPrivate::TSubmenuPopup*>(last_launching_item->fMenu->Window());
						if ( father && !father->IsSubmenu(target) )
							father->CloseAllSubMenus();
					}

					item->fMenu->fHighlightedItem = item;
					item->fMenu->fHighlightedItemIndex = item->fMenu->IndexOf( item );

					if ( target->Lock() )
					{
						item->Highlight( true );
						target->Unlock();
					}

					last_valid_target = target;
				}

				phitem = item;
				prev_target = target;
			}
			else
				times_in_this_item ++;

			if ( times_in_this_item >= 25 ) // 250000ms
			{
				if ( item && item->fSubmenu && !item->fSubmenu->Window() )
				{
					last_launching_item = item;
					target->Lock();
					BPoint point2( target->Bounds().Width() - 5, item->fFrame.top - 5 );
					target->ConvertToScreen( &point2 );
					target->Unlock();
					target->fSubmenu = new TPrivate::TSubmenuPopup( item->fSubmenu, point2 );
					target->fSubmenu->fParent = target;
					item->fSubmenu->SetUseAlphaBlending( target->fMenu->UseAlphaBlending() );
					target->fSubmenu->Go();
				}
			}

		}
		snooze( 10000 );
	}

	return 0;
}

// todo: it is possible for the whole menu not being reconstructed but just a part of it
void TMenu::InvalidateSizes( TMenuItem *item )
{
	float maxwidth = 125;
	int32 count = fItems.CountItems();
	float top = fItemMargins.top;

	for ( int32 index=0; index<count; index++ )
	{
		item = static_cast<TMenuItem*>( fItems.ItemAt(index) );
		float w, h;
		item->GetContentSize( &w, &h );
		if ( w > maxwidth )
			maxwidth = w;
	}

	if ( maxwidth > fMaxContentWidth )
		maxwidth = fMaxContentWidth;

	for ( int32 index=0; index<count; index++ )
	{
		item = static_cast<TMenuItem*>( fItems.ItemAt(index) );
		float w, h;
		item->GetContentSize( &w, &h );
		item->fFrame.left = fItemMargins.left;
		item->fFrame.right = item->fFrame.left + maxwidth;
		item->fFrame.top = top;
		item->fFrame.bottom = item->fFrame.top + h;
		top = item->fFrame.bottom;
	}

	maxwidth += fItemMargins.left + fItemMargins.right;
	top += fItemMargins.bottom;

	ResizeTo( maxwidth, top );
}

void TMenu::DeliverMessage( TMenuItem *item )
{
	if ( item && item->fMessage )
	{
		BMessage message( *(item->fMessage) );
		item->Invoke( &message );
		fTarget.SendMessage( &message );
	}
	else if ( item )
		item->Invoke(0);
}

// todo: optimize this, maybe by "divide and conquer" :-)
TMenuItem *TMenu::ItemAtPoint( BPoint where )
{
	int32 start, end;

	if ( !fHighlightedItem )
	{
		start = 0;
		end = fItems.CountItems();
	}
	else
	{
		if ( where.y > fHighlightedItem->fFrame.bottom )
		{
			start = fHighlightedItemIndex + 1;
			end = fItems.CountItems();
		}
		else
		{
			start = 0;
			end = fHighlightedItemIndex+1;
		}
	}

	for ( start; start<end; start ++ )
	{
		TMenuItem *item = static_cast<TMenuItem*>(fItems.ItemAt(start));
		if ( item->fFrame.Contains( where) )
		{
			return item;
		}
	}

	return 0;
}

// alpha blending stuff
void TMenu::PrepareAlphaBlending()
{
//	if ( fUseAlphaBlending )
//	{
		BRect r = Window()->Frame();
		r.right ++;

		BScreen screen( B_MAIN_SCREEN_ID );

//		hm..
//		if ( r.left < 0 ) r.left = 0;
//		if ( r.right > screen.Frame().Width() ) r.right = screen.Frame().Width();
//		if ( r.top < 0 ) r.top = 0;
//		if ( r.bottom > screen.Frame().Height() ) r.bottom = screen.Frame().Height();

		if ( fAlphedBack )
		{
			delete fAlphedBack;
			fAlphedBack = 0;
		}

		if ( screen.GetBitmap( &fAlphedBack, false, &r) != B_OK )
		{
			fUseAlphaBlending = false;
		}
//	}
}

void TMenu::SetUseAlphaBlending( bool b )
{
	fUseAlphaBlending = b;
}

bool TMenu::UseAlphaBlending() const
{
	return fUseAlphaBlending;
}

TPopupMenu::TPopupMenu( const char *label, bool radioMode, bool labelFromMarked )
	: TMenu( label ), fAsyncDestruct(false)
{
	fThread = -1;

	SetRadioMode( radioMode );
	SetLabelFromMarked( labelFromMarked );
}

TPopupMenu::~TPopupMenu()
{
}

BPoint TPopupMenu::ScreenLocation()
{
	return Window()->Frame().LeftTop();
}

namespace TPrivate {

struct _TPopupMenu_Go_Arguments
{
	TPopupMenu *popupmenu;;
	BPoint screenPoint;
	bool deliversMessage;
	bool openAnyway;
};

};

TMenuItem *TPopupMenu::Go( BPoint screenPoint, bool deliversMessage, bool openAnyway, bool asynchronous )
{
	if ( asynchronous )
	{
		TPrivate::_TPopupMenu_Go_Arguments *args = new TPrivate::_TPopupMenu_Go_Arguments;
		args->popupmenu = this;
		args->screenPoint = screenPoint;
		args->deliversMessage = deliversMessage;
		args->openAnyway = openAnyway;
		fThread = spawn_thread( _thread, Name(), B_NORMAL_PRIORITY, args );
		resume_thread( fThread );
		return 0;
	}
	else
	{
		return _go( screenPoint, deliversMessage, openAnyway );
	}
}

void TPopupMenu::SetAsyncAutoDestruct( bool state )
{
	fAsyncDestruct = state;
}

bool TPopupMenu::AsyncAutoDestruct() const
{
	return fAsyncDestruct;
}

TMenuItem *TPopupMenu::_go( BPoint screenPoint, bool deliversMessage, bool openAnyway )
{
	InvalidateSizes();
	if ( !Window() )
	{
		TPrivate::TSubmenuPopup *window = new TPrivate::TSubmenuPopup( this, screenPoint );
		window->fParent = 0;
		window->Go(false);
	}

	TMenuItem *item = Track( openAnyway );

	if ( deliversMessage )
		DeliverMessage( item );

	if ( fAsyncDestruct )
	{
		Window()->Lock();
		Window()->Quit();
		delete this;
		return item;
	}
	else
		Window()->Hide();

	return item;
}

int32 TPopupMenu::_thread( void *pargs )
{
	TPrivate::_TPopupMenu_Go_Arguments *args = static_cast<TPrivate::_TPopupMenu_Go_Arguments*>(pargs);
	args->popupmenu->_go( args->screenPoint, args->deliversMessage, args->openAnyway );
	delete args;
	return 0;
}
