#ifndef NMENU_H
#define NMENU_H

#include <Point.h>
#include <Rect.h>
#include <View.h>
#include <String.h>
#include <List.h>
#include <Messenger.h>

class BMessage;
class BHandler;
class BBitmap;
class BWindow;

class TMenu;

// general menu item [follows BMenuItem, check BeBook for details]
class TMenuItem
{
public:
	TMenuItem( const char *label, BMessage * = 0);
	TMenuItem( TMenu *submenu, BMessage * = 0 );
	virtual ~TMenuItem();

	BPoint ContentLocation() const;

	virtual void Draw();
	virtual void DrawContent();

	BRect Frame() const;
	virtual void GetContentSize( float *width, float *height );

	virtual void Highlight( bool flag );

	virtual status_t Invoke( BMessage *message = NULL );

	TMenu *Menu() const;

	bool IsSelected() const;

	virtual void SetEnabled( bool enabled );
	bool IsEnabled() const;

	virtual void SetLabel( const char *string );
	const char *Label() const;

	virtual void SetMarked( bool flag );
	bool IsMarked() const;

	// <unimplemented>
	virtual void SetShortcut( char shortcut, uint32 modifiers );
	char Shortcut( uint32 *modifiers = NULL ) const;

	virtual void SetTrigger( char trigger );
	char Trigger() const;
	// </unimplemented>

	TMenu *Submenu() const;

	virtual void TruncateLabel( float maxWidth, char *newLabel );

protected:
friend class TMenu;
	virtual BMessage *DragMessage();
	void DrawHighlightFrame();

	// <unimplemented>
	bool fDraggable;
	// </unimplemented>

private:
	BRect fFrame;
	TMenu *fMenu;

	BString fLabel;
	BMessage *fMessage;
	TMenu *fSubmenu;
	bool fMarked;
	bool fEnabled;
	bool fSelected;

};

class TSeparatorItem : public TMenuItem
{
public:
	TSeparatorItem();

	virtual void SetEnabled( bool state );
	virtual void GetContentSize( float *width, float *height );

	virtual void Draw();
};

class TBitmapMenuItem : public TMenuItem
{
public:
	TBitmapMenuItem( const char *label, BBitmap *bitmap, BMessage * = 0);
	TBitmapMenuItem( TMenu *submenu, BBitmap *bitmap, BMessage * = 0);
	virtual ~TBitmapMenuItem();

	virtual void GetContentSize( float *width, float *height );
	virtual void DrawContent();

	virtual void TruncateLabel( float maxWidth, char *newLabel );

	void SetBitmapAutoDestruct( bool state );
	bool BitmapAutoDestruct() const;

protected:
	BBitmap *fBitmap;

private:
	bool fBitmapAutoDestruct;
};

namespace TPrivate
{
	// internal class, a BWindow that is the parent of all popups
	class TSubmenuPopup;
};

class TMenu : public BView
{
public:
	TMenu( const char *name );
	virtual ~TMenu();

	bool AddItem( TMenuItem *item );
	bool AddItem( TMenuItem *item, int32 index );
	bool AddItem( TMenu *submenu );
	bool AddItem( TMenu *submenu, int32 index );

	bool AddSeparatorItem();

	virtual void AttachedToWindow();

	int32 CountItems() const;

	virtual void Draw( BRect updateRect );

	TMenuItem *FindItem( const char *label ) const;
	TMenuItem *FindItem( uint32 command ) const;

	TMenuItem *FindMarked();

	void Hide();
	void Show( bool selectFirst );
	virtual void Show();

	int32 IndexOf( TMenuItem *item ) const;
	int32 IndexOf( TMenu *submenu ) const;

	void InvalidateLayout();

	TMenuItem *ItemAt( int32 index ) const;
	TMenu *SubmenuAt( int32 index ) const;

	// <unimplemented>
	virtual void KeyDown( const char *bytes, int32 numBytes );
	// </unimplemented>

	TMenuItem *RemoveItem( int32 index );
	bool RemoveItem( TMenuItem *item );
	bool RemoveItem( TMenu *submenu );

	virtual BPoint ScreenLocation();

	virtual void SetEnabled( bool enabled );
	bool IsEnabled() const;

	void SetItemMargins( float left, float top, float right, float bottom );
	void GetItemMargins( float *left, float *top, float *right, float *bottom );

	// <unimplemented>
	void SetLabelFromMarked( bool flag );
	bool IsLabelFromMarked();
	// </unimplemented>

	virtual void SetMaxContentWidth( float width );
	float MaxContentWidth() const;

	virtual void SetRadioMode( bool flag );
	bool IsRadioMode();

	virtual status_t SetTargetForItems( BHandler *handler );
	virtual status_t SetTargetForItems( BMessenger messenger );

	// <unimplemented>
	virtual void SetTriggersEnabled( bool flag );
	bool AreTriggersEnabled() const;
	// </unimplemented>

	// <unimplemented>
	TMenuItem *Superitem() const;
	TMenu *Supermenu() const;
	// </unimplemented>

	TMenuItem *Track( bool openAnyway = false, BRect *clickToOpenRect = NULL );

	void InvalidateSizes( TMenuItem *item = 0 );

	void SetUseAlphaBlending( bool b );
	bool UseAlphaBlending() const;

protected:
	void DeliverMessage( TMenuItem *item );

private:
	friend class TPrivate::TSubmenuPopup;
	TMenuItem *ItemAtPoint( BPoint where );
	void PrepareAlphaBlending();

	BList fItems;

	BRect fItemMargins;
	float fMaxContentWidth;

	TMenuItem *fHighlightedItem;
	int32 fHighlightedItemIndex;

	rgb_color fBackgroundColor;
	BBitmap *fAlphedBack;
	bool fUseAlphaBlending;

	bool fRadioMode;
	bool fEnabled;

	BMessenger fTarget;

	TMenuItem *fSuperItem;
};

class TPopupMenu : public TMenu
{
public:
	TPopupMenu( const char *label, bool radioMode = true, bool labelFromMarked = true );
	virtual ~TPopupMenu();

	virtual BPoint ScreenLocation();

	TMenuItem *Go( BPoint screenPoint, bool deliversMessage = false, bool openAnyway = false, bool asynchronous = false );

	void SetAsyncAutoDestruct( bool state );
	bool AsyncAutoDestruct() const;

private:
	TMenuItem *_go( BPoint screenPoint, bool deliversMessage, bool openAnyway );
	static int32 _thread( void * );

	bool fAsyncDestruct;

	thread_id fThread;
};

#endif /* NMENU_H */
